# (c) 2013-2018 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""This module contains the ImageCache, caching formulas which have already been
converted. This allows to re-use images for formulas which occur multiple timesd
within a document. Furthermore, it can significantly speed up incremental
document creation, because the cache is remembered across GladTeX runs.

Cache format:

    { # dict of formulas
        'some formula': # formula as key into dictionary
            { # list of display math / inline maths variants
                True: # displaymath = True
                    { # dictionary of values describing formula
                        'path': 'some/path'
                        'pos': { # positioning within the HTML document
                            'height': ..., 'width':..., 'depth:....
                        }
                    }
                    }
            }
    }

Formulas are `normalized`, so spacing is unified to detect possibly equal
formulas more easyly.
"""

import contextlib
import json
import os

CACHE_VERSION = '2.0'

def normalize_formula(formula):
    """This function normalizes a formula. This e.g. means that multiple white
    spaces are squeezed into one and a tab will be replaced by a space. With
    this it is more realistic that a recurring formula in a document is detected
    as such, even though if it might have been written with different spacing.
    Empty braces ({}) are removed as well."""
    return formula.replace('{}', ' ').replace('\t', ' ').replace('  ', ' '). \
        rstrip().lstrip()

def recover_bools(object):
    """After JSon is read from disk, keys as False or True have been serialized
    to 'false' and 'true', but they're not recovered by the json parser. This
    function alters converts these keys back to booleans; note: it only works
    with references, so this function doesn't return anything."""
    if isinstance(object, dict):
        for key in ['false', 'true']:
            if key in object:
                val = object[key] # store value
                object[key == 'true'] = val # safe it with boolean representation
                del object[key] # remove string key
        # iterate recursively through dict
        for value in object.values():
            recover_bools(value)
    if isinstance(object, list):
        for item in object:
            recover_bools(item)

class JsonParserException(Exception):
    """Specialized exception class for handling errors while parsing the JSON
    cache."""
    pass

class ImageCache:
    """
    This cache stores formulas which have been converted already and don't need
    to be converted again. This is both a disk usage and performance
    improvement. The cache can be written and read from disk.

    If the argument keep_old_cache is True, the cache will raise a
    JsonParserException if that file could not be read (i.e. incompatible
    GladTeX version). If set to False, it'll discard the cache along with all
    eqn* files and start with a clean cache.

    Example:

    cache = ImageCache()
    c.add_formula('\\tau', # the formulas
        {'height': 1, 'depth': 2, 'width='3'}, # the positioning information for the output document
        'eqn042.svg', displaymath=True):
    assert len(cache) == 1 # one entry
    c.write()
    assert os.path.exists('gladtex.cache')

    The optional argument base_path adds the ability to add a base directory to
    each file path. The base_path is used to simulate a different working
    directory. Imagine you have a directory chapter01 and you want your images
    to be in a subdirectory img. Chjanging the current working directory isn't
    possible because of parallelism, therefore you initialise the cache like
    this:

    c = cache = ImageCache(path='/img/gladtex.cache', base_path='chapter01')
    c.add_formula(…, 'img/eqn001.svg') # will result in chapter01/img/eqn001.svg
    """
    VERSION_STR = 'GladTeX__cache__version'

    def __init__(self, path='gladtex.cache', keep_old_cache=True,
            base_path=''):
        self.__cache = {}
        self.__set_version(CACHE_VERSION)
        self.__cache_name = os.path.join(base_path, path)
        self.__base_path = base_path
        if os.path.exists(os.path.join(base_path, path)):
            try:
                self._read()
            except JsonParserException:
                if keep_old_cache:
                    raise
                else:
                    self._remove_old_cache_and_files()

    def __len__(self):
        """Return number of formulas in the cache."""
        # ignore version
        return len(self.__cache) - 1

    def __set_version(self, version):
        """Set version of cache (data structure format)."""
        self.__cache[ImageCache.VERSION_STR] = version

    def write(self):
        """Write cache to disk. The file name will be the one configured during
        initialisation of the cache."""
        if not self.__cache:
            return
        with open(self.__cache_name, 'w', encoding='UTF-8') as file:
            file.write(json.dumps(self.__cache))

    def _read(self):
        """Read Json from disk into cache, if file exists.
        :raises JsonParserException if json could not be parsed"""
        def raise_error(msg):
            raise JsonParserException(msg + "\nPlease delete the cache (and" + \
                        " the images) and rerun the program.")
        if os.path.exists(self.__cache_name):
            #pylint: disable=broad-except
            try:
                with open(self.__cache_name) as file:
                    self.__cache = json.load(file)
            except Exception as e:
                msg = "error while reading cache from %s: " % \
                        os.path.abspath(self.__cache_name)
                if isinstance(e, (ValueError, OSError)):
                    msg += str(e.args[0])
                elif isinstance(e, UnicodeDecodeError):
                    msg += 'expected UTF-8 encoding, erroneous byte ' + \
                            '{0} at {1}:{2} ({3})'.format(*(e.args[1:]))
                else:
                    msg += str(e.args[0])
                raise_error(msg)
        if not isinstance(self.__cache, dict):
            raise_error("Decoded Json is not a dictionary.")
        if not self.__cache.get(ImageCache.VERSION_STR):
            self.__set_version(CACHE_VERSION)
        cur_version = self.__cache.get(ImageCache.VERSION_STR)
        if cur_version != CACHE_VERSION:
            raise_error("Cache in %s has version %s, expected %s." % \
                    (self.__cache_name, cur_version, CACHE_VERSION))
        recover_bools(self.__cache)

    def _remove_old_cache_and_files(self):
        os.remove(self.__cache_name)
        directory = os.path.dirname(self.__cache_name)
        if not directory:
            directory = '.'
        # remove all files starting with eqn*
        for file in os.listdir(directory):
            if not file.startswith('eqn'):
                continue
            file = os.path.join(directory, file)
            if os.path.isfile(file):
                os.remove(file)

    def add_formula(self, formula, pos, file_path, displaymath=False):
        """Add formula to cache. The pos argument contains the positioning
        info for the output document and is a dict with 'height', 'width' and
        'depth'.
        Keep in mind that formulas set with displaymath are not the same as
        those set iwth inlinemath.
        This method raises OSError if specified image doesn't exist or if it got
        an absolute file_path.
        """
        if os.path.isabs(file_path):
            raise OSError(("The file path to the image may NOT be an absolute "
                    "path: ") + file_path)
        if '\\' in file_path:
            file_path = file_path.replace('\\', '/')
        if not os.path.exists(os.path.join(self.__base_path, file_path)):
            raise OSError("cannot add %s to the cache: doesn't exist" %
                    os.path.join(self.__base_path, file_path))
        if not pos or not formula or not file_path:
            raise ValueError("the supplied arguments may not be empty/none")
        if not isinstance(displaymath, bool):
            raise ValueError("displaymath must be a boolean")
        formula = normalize_formula(formula)
        if not formula in self.__cache:
            self.__cache[formula] = {}
        val = self.__cache[formula]
        if not displaymath in val:
            val[displaymath] = {'pos': pos,
                    'path': file_path,
                }

    def remove_formula(self, formula, displaymath):
        """This method removes the given formula from the cache. A KeyError is
        raised, if the formula did not exist. Internally, formulas are
        normalized to detect similarities."""
        formula = normalize_formula(formula)
        if not formula in self.__cache:
            raise KeyError("key %s not in cache" % formula)
        else:
            value = self.__cache[formula]
            if displaymath in value:
                with contextlib.suppress(FileNotFoundError):
                    os.remove(os.path.join(self.__base_path, value[displaymath]['path']))
                del self.__cache[formula][displaymath]
                if not self.__cache[formula]:
                    del self.__cache[formula]
            else:
                raise KeyError("key %s (%s) not in cache" % (formula, displaymath))

    def contains(self, formula, displaymath):
        """Check whether a formula was already cached and return True if
        found."""
        try:
            return bool(self.get_data_for(formula, displaymath))
        except KeyError:
            return False


    def get_data_for(self, formula, displaymath):
        """
        Retrieve meta data about a formula from the cache.

        The meta information is used to embed the formula in the HTML document.
        It is a dictionary with the keys 'pos' and 'path'. The positioning info
        is described in the documentation of this class.
        This method raises a KeyError if the formula wasn't found."""
        formula = normalize_formula(formula)
        if not formula in self.__cache:
            raise KeyError(formula, displaymath)
        else:
            # check whether file still exists
            value = self.__cache[formula]
            if displaymath in value.keys():
                # if file doesn't exist anymore, outdated and hence removed from
                # cache
                if not os.path.exists(os.path.join(self.__base_path,
                        value[displaymath]['path'])):
                    del self.__cache[formula]
                    raise KeyError((formula, displaymath))
                else:
                    return value[displaymath]
            else:
                raise KeyError((formula, displaymath))



# (c) 2013-2018 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""In order to convert images only if they are not already cached, the cached
converter sacrifices customizability for convenience and provides a class
converting a formula directly to a png file."""

import concurrent.futures
import multiprocessing
import os
import subprocess

from . import caching, image, typesetting
from .caching import normalize_formula
from .image import Format

class ConversionException(Exception):
    """This exception is raised whenever a problem occurs during conversion.
    Example:
    c = ConversionException("cause", "\\tau", 10, 38, 5)
    assert c.cause == cause
    assert c.formula == '\\tau'
    assert c.src_line_number == 10 # line number in source document (counting from 1)
    assert c.src_pos_on_line == 38 # position of formula in source line, counting from 1
    assert c.formula_count == 5 # fifth formula in document (starting from 1)
    """
    # mind your own business mr. pylint:
    #pylint: disable=too-many-arguments
    def __init__(self, cause, formula, formula_count, src_line_number=None,
            src_pos_on_line=None):
        # provide a default error message
        if src_line_number and src_pos_on_line:
            super().__init__("LaTeX failed at formula line {}, {}, no. {}: {}".format(
                src_line_number, src_pos_on_line, formula_count, cause))
        else:
            super().__init__("LaTeX failed at formula no. {}: {}".format(
                formula_count, cause))
        # provide attributes for upper level error handling
        self.cause = cause
        self.formula = formula
        self.src_line_number = src_line_number
        self.src_pos_on_line = src_pos_on_line
        self.formula_count = formula_count

class CachedConverter:
    """Convert formulas to images.

    c = CachedConverter(base_path)
    for formula in [... formulas ...]:
        pos, file_path = c.convert(formula)
        ...

    The formula is either converted  or retrieved from a cache in the same
    directory like the images.

    :param base_path directory of the output HTML file; link references in the
            HTML document will link relative to it
    :param keep_old_cache If an existing cache cannot be read (incompatible
        GladTeX version, ...) Aand the flag is set, the program will simply
        crash and tell the user to remove the cache (default). If set to False,
        the program will instead remove the cache and all eqn* files and
        recreate the cache.
    :param encoding The encoding for the LaTeX document, default None
    :param img_dir directory for images (default ., equivalent to base_path)
            For example "images" would put it in `base_path`/images and "../img"
            would put it in "base_path/../img"
    """
    GLADTEX_CACHE_FILE_NAME = 'gladtex.cache'

    def __init__(self, base_path, keep_old_cache=True, encoding=None,
            img_dir=''):
        empty_path = lambda p: ('' if not p or p.strip(os.sep) == '.' else p)
        self.__output_path = empty_path(base_path) # path where converted document will be
        self.__img_dir = empty_path(img_dir) # relative to base_path
        # cache path is **relative** to base_path
        cache_path = os.path.join(self.__img_dir,
                CachedConverter.GLADTEX_CACHE_FILE_NAME)
        self.__cache = caching.ImageCache(cache_path,
                keep_old_cache=keep_old_cache,
                base_path=empty_path(self.__output_path))
        self.__converter = None
        self.__options = {'dpi': None, 'transparency': None, 'fontsize': None,
                'background_color': None, 'foreground_color': None,
                'preamble': None, 'latex_maths_env': None,
                'keep_latex_source': False, 'png': False}
        self.__encoding = encoding
        self.__replace_nonascii = False


    def set_option(self, option, value):
        """Set one of the options accepted for gleetex.image.Tex2img. It is a
        proxy function.
        `option` must be one of dpi, fontsize, transparency, background_color,
        foreground_color, preamble, latex_maths_env, keep_latex_source, png."""
        if not option in self.__options.keys():
            raise ValueError("Option must be one of " + \
                    ', '.join(self.__options.keys()))
        self.__options[option] = value

    def set_replace_nonascii(self, flag):
        """If set, GladTeX will convert all non-ascii character to LaTeX
        commands. This setting is passed through to typesetting.LaTeXDocument."""
        self.__replace_nonascii = flag


    def convert_all(self, formulas):
        """convert_all(formulas)
        Convert all formulas using self.convert concurrently. Each element of
        `formulas` must be a tuple containing (formula, displaymath,
        Formulas already contained in the cache are not converted."""
        formulas_to_convert = self._get_formulas_to_convert(formulas)
        if formulas_to_convert:
            self.__converter = image.Tex2img(Format.Png
                    if self.__options['png'] else Format.Svg)
            # apply configured image output options
            for option, value in self.__options.items():
                if value and hasattr(self.__converter, 'set_' + option):
                    if isinstance(value, str): # only try string -> number
                        try: # some values are numbers
                            value = float(value)
                        except ValueError:
                            pass
                    getattr(self.__converter, 'set_' + option)(value)
            self._convert_concurrently(formulas_to_convert)

    def _get_formulas_to_convert(self, formulas):
        """Return a list of formulas to convert, along with their count in the
        global list of formulas of the document being converted and the file
        name. Function was decomposed for better testability."""
        formulas_to_convert = [] # find as many file names as equations
        file_ext = (Format.Png.value if self.__options['png']
                else Format.Svg.value)
        eqn_path = lambda x: os.path.join(self.__img_dir,
                'eqn%03d.%s' % (x, file_ext))
        abs_eqn_path = lambda x: os.path.join(self.__img_dir, eqn_path(x))

        # is (formula, display_math) already in the list of formulas to convert;
        # displaymath is important since formulas look different in inline maths
        formula_was_converted = lambda f, dsp: (normalize_formula(f), dsp) in \
                ((normalize_formula(u[0]), u[3]) for u in formulas_to_convert)
        # find enough free file names
        file_name_count = 0
        used_file_names = [] # track which file names have been assigned
        for formula_count, (pos, dsp, formula) in enumerate(formulas):
            # ToDo: this belongs in the cache
            if not self.__cache.contains(formula, dsp) and \
                    not formula_was_converted(formula, dsp):
                while os.path.exists(abs_eqn_path(file_name_count)) or \
                    eqn_path(file_name_count) in used_file_names:
                    file_name_count += 1
                used_file_names.append(eqn_path(file_name_count))
                formulas_to_convert.append((formula, pos, eqn_path(file_name_count),
                    dsp, formula_count + 1))
        return formulas_to_convert


    def _convert_concurrently(self, formulas_to_convert):
        """The actual concurrent conversion process. Method is intended to be
        called from convert_all()."""
        imgdir_full = os.path.join(self.__output_path, self.__img_dir)
        if imgdir_full and not os.path.exists(imgdir_full):
            # create directory *before* it is required in the concurrent
            # formulacreation step
            os.makedirs(imgdir_full)

        thread_count = int(multiprocessing.cpu_count() * 2.5)
        # convert missing formulas
        with concurrent.futures.ThreadPoolExecutor(max_workers=thread_count) as executor:
            # start conversion and mark each thread with its formula, position
            # in the source file and formula_count (index into a global list of
            # formulas)
            jobs = {executor.submit(self.__convert, eqn, path, dsp): (eqn, pos, count)
                for (eqn, pos, path, dsp, count) in formulas_to_convert}
            error_occurred = None
            for future in concurrent.futures.as_completed(jobs):
                if error_occurred and not future.done():
                    future.cancel()
                    continue
                formula, pos_in_src, formula_count = jobs[future]
                try:
                    data = future.result()
                except subprocess.SubprocessError as e:
                    # retrieve the position (line, pos on line) in the source
                    # document from original formula list
                    if pos_in_src: # missing for the pandocfilter case
                        pos_in_src = list(p+1 for p in pos_in_src) # user expects lines/pos_in_src' to count from 1
                    self.__cache.write() # write back cache with valid entries
                    if not pos_in_src: # pandocfilter case:
                        error_occurred = ConversionException(str(e.args[0]),
                                formula, formula_count)
                    else:
                        error_occurred = ConversionException(str(e.args[0]), formula,
                            formula_count, pos_in_src[0], pos_in_src[1])
                else:
                    self.__cache.add_formula(formula, data['pos'],
                            data['path'],
                            data['displaymath'])
                    self.__cache.write()
            #pylint: disable=raising-bad-type
            if error_occurred:
                raise error_occurred



    def __convert(self, formula, img_path, displaymath=False):
        """convert(formula, img_path, displaymath=False)
        Convert given formula with displaymath/inlinemath.
        This method wraps the formula in a tex document, executes all the steps
        to produce a image and return the positioning information for the
        HTML output. It does not check the cache.
        :param formula formula to convert
        :param img_path image output path (relative to the configured base_path,
                    see __init__)
        :param displaymath whether or not to use displaymath during the conversion
        :return dictionary with position (pos), image path (path) and formula
            style (displaymath, boolean) as a dictionary with the keys in
            parenthesis"""
        latex = typesetting.LaTeXDocument(formula)
        latex.set_displaymath(displaymath)
        def set(opt, setter):
            if self.__options[opt]:
                getattr(latex, 'set_' + setter)(self.__options[opt])
        set('preamble', 'preamble_string')
        set('latex_maths_env', 'latex_environment')
        set('background_color', 'background_color')
        set('foreground_color', 'foreground_color')
        if self.__encoding:
            latex.set_encoding(self.__encoding)
        if self.__replace_nonascii:
            latex.set_replace_nonascii(True)
        # dvipng needs the additionalindication of transparency (enabled by
        # default) when setting a background colour
        if self.__options['background_color']:
            self.__converter.set_transparency(False)
        pos = self.__converter.convert(latex,
                os.path.join(self.__output_path, os.path.splitext(img_path)[0]))
        return {'pos': pos,
                'path': img_path, # relative to self.__base_name(!)
                'displaymath': displaymath}

    def get_data_for(self, formula, display_math):
        """Simple wrapper around ImageCache, enriching the returned data with
        the information provided as arguments to this function. This helps when
        using a formula without its context."""
        data = self.__cache.get_data_for(formula, display_math).copy()
        data.update({'formula': formula, 'displaymath': display_math})
        return data



# (c) 2013-2018 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""This module contains functionality to typeset formuas for the usage in a
LaTeX document (e.g. creating the preamble, replacing non-ascii letters) and to
typeset LaTeX formulas in a more readable way as alternate description of the
resulting image. """

from . import unicode

FORMATTING_COMMANDS = ['\\ ', '\\,', '\\;', '\\big', '\\Big', '\\left',
        '\\right', '\\limits']



class DocumentSerializationException(Exception):
    """This error is raised whenever a non-ascii character contained in a
    formula could not be replaced by a LaTeX command.
    It provides the following attributes:
    formula - the formula
    index - position in formula
    upoint - unicode point."""
    def __init__(self, formula, index, upoint):
        self.formula = formula
        self.index = index
        self.upoint = upoint
        super().__init__(formula, index, upoint)

    def __str__(self):
        return ("could not find LaTeX replacement command for unicode "
                "character %d, index %d in formula %s") % (self.upoint,
                    self.index, self.formula)


def escape_unicode_maths(formula, replace_alphabeticals=True):
    """This function uses the unicode table to replace any non-ascii character
    (identified with its unicode code point)  with a LaTeX command.
    It also parses the formula for commands as e.g. \\\text or \\mbox and
    applies text-mode commands within them.
    This allows the conversion of formulas with unicode maths with old-style
    LaTeX2e, which gleetex depends on."""
    if not any(ord(ch) > 160 for ch in formula):
        return formula # no umlauts, no replacement

    # characters in math mode need a different replacement than in text mode.
    # Therefore, the string has to be split into parts of math and text mode.
    chunks = []
    if not ('\\text' in formula or '\\mbox' in formula):
        #    no text mode, so tread a
        chunks = [formula]
    else:
        start = 0
        while '\\text' in formula[start:] or '\\mbox' in formula[start:]:
            index = formula[start:].find('\\text')
            if index < 0:
                index = formula[start:].find('\\mbox')
            opening_brace = formula[start + index:].find('{') + start + index
            # add text before text-alike command and the command itself to chunks
            chunks.append(formula[start:opening_brace])
            closing_brace = get_matching_brace(formula, opening_brace)
            # add text-mode stuff
            chunks.append(formula[opening_brace:closing_brace + 1])
            start = closing_brace + 1
        # add last chunk
        chunks.append(formula[start:])

    is_math = True
    for index, chunk in enumerate(chunks):
        try:
            chunks[index] = replace_unicode_characters(chunk, is_math,
                    replace_alphabeticals=replace_alphabeticals)
        except ValueError as e: # unicode point missing
            index = int(e.args[0])
            raise DocumentSerializationException(formula, index,
                    ord(formula[index])) from None
        is_math = not is_math
    return ''.join(chunks)


def replace_unicode_characters(characters, is_math, replace_alphabeticals=True):
    """Replace all non-ascii characters within the given string with their LaTeX
    equivalent. The boolean is_math indicates, whether text-mode commands (like
    in \\text{}) or the amsmath equivalents should be used.
    When replace_alphabeticals is False, alphabetical characters will not be
    replaced through their LaTeX command when in text mode, so that text within
    \\text{} (and similar) is not garbled. For instance, \\text{für} is be
    replaced by \\text{f\"{u}r} when replace_alphabeticals=True. This is useful
    for the alt attribute of an image, where the reader might want to read
    the normal text as such.
    This function raises a ValueError if a unicode point is not in the table.
    The first argument of the ValueError is the index within the string, where
    the unknown unicode character has been encountered."""
    result = []
    for idx, character in enumerate(characters):
        if ord(character) < 168: # ignore normal ascii character and unicode control sequences
            result.append(character)
        # treat alphanumerical characters differently when in text mode, see doc
        # string; don't replace alphabeticals if specified
        elif character.isalpha() and not replace_alphabeticals:
            result.append(character)
        else:
            mode = (unicode.LaTeXMode.mathmode if is_math else
                    unicode.LaTeXMode.textmode)
            commands = unicode.unicode_table.get(ord(character))
            if not commands: # unicode point missing in table
                # is catched one level above; provide index for more concise error output
                raise ValueError(characters.index(character))
            # if math mode and only a text alternative exists, add \\text{}
            # around it
            if mode == unicode.LaTeXMode.mathmode and mode not in commands:
                result.append('\\text{%s}' % commands[unicode.LaTeXMode.textmode])
            else:
                result.append(commands[mode])
                # if the next character is alphabetical, add space
                if (idx+1) < len(characters) and characters[idx+1].isalpha() \
                        and commands[mode][-1].isalpha():
                    result.append(' ')
    return ''.join(result)

def get_matching_brace(string, pos_of_opening_brace):
    if string[pos_of_opening_brace] != '{':
        raise ValueError("index %s in string %s: not a opening brace" % \
            (pos_of_opening_brace, repr(string)))
    counter = 1
    for index, ch in enumerate(string[pos_of_opening_brace + 1:]):
        if ch == '{':
            counter += 1
        elif ch == '}':
            counter -= 1
            if counter == 0:
                return pos_of_opening_brace + index + 1
    if counter != 0:
        raise ValueError("Unbalanced braces in formula " + repr(string))



#pylint: disable=too-many-instance-attributes
class LaTeXDocument:
    """This class represents a LaTeX document. It is intended to contain an
    equation as main content and properties to customize it. Its main purpose is
    to provide a str method which will serialize it to a full LaTeX document.
    """
    def __init__(self, eqn):
        self.__encoding = None
        self.__equation = eqn
        self.__displaymath = False
        self.__fontsize = 12
        self.__background_color = None
        self.__foreground_color = None
        self._preamble = ''
        self.__maths_env = None
        self.__replace_nonascii = False

    def _parse_color(self, color):
        # could be a valid color name
        try: # hex number?
            return int(color, 16)
        except ValueError:
            return color # treat as normal dvips compatible colour name

    def set_background_color(self, color):
        """Set the background color. The `color` can be either a valid dvips
        name or a tuple with RGB values between 0 and 1. If unset, the image
        will be transparent."""
        self.__background_color = self._parse_color(color)

    def set_foreground_color(self, color):
        """Set the foreground color. The `color` can be either a valid dvips
        name or a tuple with RGB values between 0 and 1. If unset, the text
        will be black."""
        self.__foreground_color = self._parse_color(color)

    def set_replace_nonascii(self, flag):
        """If True, all non-ascii character will be replaced through a LaTeX
        command."""
        self.__replace_nonascii = flag

    def set_latex_environment(self, env):
        """Set maths environment name like `displaymath` or `flalign*`."""
        self.__maths_env = env

    def get_latex_environment(self):
        return self.__maths_env

    def get_encoding(self):
        """Return encoding for the document (or None)."""
        return self.__encoding

    def set_preamble_string(self, p):
        """Set the string to add to the preamble of the LaTeX document."""
        self._preamble = p

    def set_encoding(self, encoding):
        """Set the encoding as used by the inputenc package."""
        if encoding.lower().startswith('utf') and '8' in encoding:
            self.__encoding = 'utf8'
        elif (encoding.lower().startswith('iso') and '8859' in encoding) or \
                encoding.lower() == 'latin1':
            self.__encoding = 'latin1'
        else:
            # if you plan to add an encoding, you have to adjust the str
            # function, which also loads the  fontenc package
            raise ValueError(("Encoding %s is not supported at the moment. If "
                "you want to use LaTeX 2e, you should report a bug at the home "
                "page of GladTeX.") % encoding)

    def set_displaymath(self, flag):
        """Set whether the formula is set in displaymath."""
        if not isinstance(flag, bool):
            raise TypeError("Displaymath parameter must be of type bool.")
        self.__displaymath = flag

    def is_displaymath(self):
        return self.__displaymath

    def _get_encoding_preamble(self):
        # first check whether there are umlauts within the formula and if so, an
        # encoding has been set
        if any(ord(ch) > 128 for ch in self.__equation) and \
                not self.__replace_nonascii:
            if not self.__encoding:
                raise ValueError(("No encoding set, but non-ascii characters "
                        "present. Please specify an encoding."))
        encoding_preamble = ''
        if self.__encoding:
            # try to guess language and hence character set (fontenc)
            import locale
            language = locale.getdefaultlocale()
            if language and language[0]: # extract just the language code
                language = language[0].split('_')[0]
            if not language or not language[0]:
                language = 'en'
            # check whether language on computer is within T1 and hence whether
            # it should be loaded; I know that this can be a misleading
            # assumption, but there's no better way that I know of
            if language in ['fr', 'es', 'it', 'de', 'nl', 'ro', 'en']:
                encoding_preamble += '\n\\usepackage[T1]{fontenc}'
            else:
                raise ValueError(("Language not supported by T1 fontenc "
                    "encoding; please report this to the GladTeX project."))
        return encoding_preamble

    def set_fontsize(self, size_in_pt):
        """Set fontsize in pt, 12 pt by default."""
        self.__fontsize = size_in_pt

    def get_fontsize(self, size_in_pt):
        return self.__fontsize

    def __str__(self):
        preamble = self._get_encoding_preamble() + \
                ('\n\\usepackage[utf8]{inputenc}\n\\usepackage{amsmath, amssymb}'
                '\n') + (self._preamble if self._preamble else '')
        return self._format_document(preamble)

    def _format_color_definition(self, which):
        color = getattr(self, '_%s__%s_color' % (self.__class__.__name__,
            which))
        if not color or isinstance(color, str):
            return ''
        return ('\\definecolor{%s}{HTML}{%s}' % (which,
                    hex(color)[2:].upper().zfill(6)))

    def _format_colors(self):
        color_defs = (self._format_color_definition('background'),
                self._format_color_definition('foreground'),)
        color_body = ''
        if self.__background_color:
            color_body += ('\\pagecolor{%s}' % ('background' if color_defs[0]
                        else self.__background_color))
        if self.__foreground_color:
            # opening brace isn't required here, inserted automatically
            color_body += ('\\color{%s}' % ('foreground' if color_defs[1]
                    else self.__foreground_color))
        return (''.join(color_defs), color_body)


    def _format_document(self, preamble):
        """Return a formatted LaTeX document with the specified formula
        embedded."""
        opening, closing = None,None
        if self.__maths_env:
            opening = '\\begin{%s}' % self.__maths_env
            closing = '\\end{%s}' % self.__maths_env
        else:
            # determine characters with which to surround the formula
            opening = '\\[' if self.__displaymath else '\\('
            closing = '\\]' if self.__displaymath else '\\)'
        formula = self.__equation.lstrip().rstrip()
        if self.__replace_nonascii:
            formula = escape_unicode_maths(formula, replace_alphabeticals=True)
        fontsize = 'fontsize=%ipt' % self.__fontsize
        color_preamble, color_body = self._format_colors()
        return ("\\documentclass[%s, fleqn]{scrartcl}\n\n%s\n"
            "\\usepackage[dvipsnames]{xcolor}\n"
            "%s\n" # color definitions, if applicable
            "\\usepackage[active,textmath,displaymath,tightpage]{preview} "
            "%% must be last one, see doc\n\n\\begin{document}\n"
            "\\noindent%%\n"
            "\\begin{preview}{%s"
            "%s%s%s}\\end{preview}\n"
            "\\end{document}\n") % (fontsize, preamble, color_preamble,
                    color_body, opening, formula, closing)


def increase_readability(formula, replace_nonascii=False):
    """In alternate texts for non-image users or those using a screen reader,
    the LaTeX code should be as readable as possible. Therefore the formula
    should not contain unicode characters or formatting instructions."""
    if replace_nonascii:
        # keep umlauts, etc; makes the alt more readable, yet wouldn't compile
        formula = escape_unicode_maths(formula, replace_alphabeticals=False)
    # replace formatting-only symbols which distract the reader
    formula_changed = True
    while formula_changed:
        formula_changed = False
        for command in FORMATTING_COMMANDS:
            idx = formula.find(command)
            # only replace if it's not after a \\ and not part of a longer command
            if (idx > 0 and formula[idx-1] != '\\') or idx == 0:
                end = idx + len(command)
                # following conditions for replacement must be met:
                # command doesn't end on alphabet. char. and is followed by same
                # category OR end of string reached OR command does not # end on
                # alphabetical char. at all
                if end >= len(formula) or not command[-1].isalpha() \
                        or not formula[end].isalpha():
                    formula = formula[:idx] + ' ' + formula[idx + len(command):]
                    formula = formula.replace('  ', ' ')
                    formula_changed = True
    return formula



# (c) 2013-2018 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""This module contains functionality to parse formulas from a given Pandoc
document AST and to replace these through formatted HTML equations. Even
though this could be done in a single run, this would conflict with the internal
GleeTeX structure and allows for an easy parallelisation of the formula
conversion."""

import json

from .htmlhandling import ParseException

def __extract_formulas(formulas, ast):
    """Recursively extract 'Math' elements from the given AST and add them to
    `formulas (list)`."""
    if isinstance(ast, list):
        for item in ast:
            __extract_formulas(formulas, item)
    elif isinstance(ast, dict):
        if 't' in ast and ast['t'] == 'Math':
            style, formula = ast['c']
            # style = {'t': 'blah'} -> we want blah
            style = next(iter(style.values()))
            if style not in ['InlineMath', 'DisplayMath']:
                raise ParseException("[pandoc] unknown formula formatting: " + \
                        repr(ast['c']))
            style = (True if style == 'DisplayMath' else False)
            # position is None (only applicable for HTML parsing)
            formulas.append((None, style, formula))
        elif 'c' in ast:
            __extract_formulas(formulas, ast['c'])
    #    ^ all other cases do not matter

def extract_formulas(ast):
    """Extract formulas from a given Pandoc document AST.
    The returned formulas are typed like those form the HTML parser, therefore
    the first argument of the tuple is unused and hence None.
    :param  ast  Structure of lists and dicts representing a Pandoc document AST
    :return a list of formulas where each formula is (None, style, formula)"""
    formulas = []
    __extract_formulas(formulas, ast['blocks'])
    return formulas

def replace_formulas_in_ast(formatter, ast, formulas):
    """replace 'Math' elements from the given AST with a formatted variant
    Each 'Math' element found in the Pandoc AST will be replaced through a
    formatted (HTML) image link. The formulas are taken from the supplied
    formulas list. The number of formulas in the document has to match the
    number of formulas form the list."""
    if not formulas:
        return
    if isinstance(ast, list):
        for item in ast:
            replace_formulas_in_ast(formatter, item, formulas)
    elif isinstance(ast, dict):
        if 't' in ast and ast['t'] == 'Math':
            ast['t'] = 'RawInline' # raw HTML
            eqn = formulas.pop(0)
            ast['c'] = ['html',formatter.format(eqn['pos'], eqn['formula'], eqn['path'],
                    eqn['displaymath'])]
        elif 'c' in ast:
            replace_formulas_in_ast(formatter, ast['c'], formulas)
    # ^ ignore all other cases

def write_pandoc_ast(file, document, formatter):
    """Replace 'Math' elements from a Pandoc AST with 'RawInline' elements,
    containing formatted HTML image tags.
    :param formatter    A formatter offering the "format" method (see ImageFormatter)
    :param formulas     A list of formulas with the information (pos, formula, path, displaymath)
    :param ast          Document ast to modified"""
    ast, formulas = document
    replace_formulas_in_ast(formatter, ast['blocks'], formulas)
    file.write(json.dumps(ast))




# (c) 2013-2019 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""GleeTeX is designed to allow the re-use of the image creation code
independently of the HTML conversion code. Therefore, this modules contains the
code required to parse equations from HTML, to write converted HTML documents
back and to handle the outsourcing of formulas too long for an HTML alt tag to
an external file.
The pandoc module contains similar functions for using GleeTeX as a pandoc
filter."""

import collections
import enum
import html.parser
import os
import posixpath
import re

from . import typesetting

# match HTML 4 and 5
CHARSET_PATTERN = re.compile(
        rb'(?:content="text/html; charset=(.*?)"|charset="(.*?)")')

class ParseException(Exception):
    """Exception to propagate a parsing error."""
    def __init__(self, msg, pos=None):
        self.msg = msg
        self.pos = pos
        super().__init__(msg, pos)

    def __str__(self):
        return ('line {0.pos[0]}, {0.pos[1]}: {0.msg}'.format(self)
                if self.pos else self.msg)

def get_position(document, index):
    """This returns the line number and position on line for the given String.
    Note: lines and positions are counted from 0."""
    line = document[:index+1].count('\n')
    if document[index] == '\n':
        return (line, 0)
    newline = document[:index+1].rfind('\n')
    newline = (newline if newline >= 0 else 0)
    return (line, len(document[newline:index]))


def find_anycase(where, what):
    """Find with both lower or upper case."""
    lower = where.find(what.lower())
    upper = where.find(what.upper())
    if lower >= 0:
        return lower
    return upper

class EqnParser:
    """This parser parses <eq>...</eq> our of a document. It's not an HTML
    parser, because the content within <eq>.*<eq> is parsed verbatim.
    It also parses comments, to not consider formulas within comments. All other
    cases are unhandled. Especially CData is problematic, although it seems like
    a rare use case."""
    class State(enum.Enum): # ([\s\S]*?) also matches newlines
        Comment = re.compile(r'<!--([\s\S]*?)-->', re.MULTILINE)
        Equation = re.compile(r'<\s*(?:eq|EQ)\s*(.*?)?>([\s\S.]+?)<\s*/\s*(?:eq|EQ)>',
                re.MULTILINE)

    HTML_ENTITY = re.compile(r'(&(:?#\d+|[a-zA-Z]+);)')

    def __init__(self):
        self.__document = None
        self.__data = []
        self.__encoding = None

    def feed(self, document):
        """Feed a string or a bytes instance and start parsing. If a bytes
        instance is fed, an HTML encoding header has to be present, so that the
        encoding can be extracted."""
        if isinstance(document, bytes): # try to guess encoding
            try:
                encoding = next(filter(bool, CHARSET_PATTERN.search(document)
                        .groups())).decode('ascii')
                document = document.decode(encoding)
            except AttributeError:
                raise ParseException(("Could not determine encoding of "
                        "document, no charset information in the HTML header "
                        "found."))
            self.__encoding = encoding
        self.__document = document[:]
        self._parse()

    def find_with_offset(self, doc, start, what):
        """This find method searches in the document for a given string, staking
        the offset into account. REturned is the absolute position (so offset +
        relative match position) or -1 for no hit."""
        if isinstance(what, str):
            pos = doc[start:].find(what)
        else:
            match = what.search(doc[start:])
            pos = (-1 if not match else match.span()[0])
        return (pos if pos == -1 else pos + start)


    def _parse(self):
        """This function parses the document, while maintaining state using the
        State enum."""
        in_document = lambda x: not x == -1
        # maintain a lower-case copy, which eases searching, but doesn't affect
        # the handler methods
        doc = self.__document[:].lower()

        end = len(self.__document) - 1
        eq_start = re.compile(r'<\s*eq\s*(.*?)>')

        start_pos = 0
        while start_pos < end:
            comment = self.find_with_offset(doc, start_pos, '<!--')
            formula = self.find_with_offset(doc, start_pos, eq_start)
            if in_document(comment) and in_document(formula): # both present, take closest
                if comment < formula:
                    self.__data.append(self.__document[start_pos:comment])
                    start_pos = self.handle_comment(comment)
                else:
                    self.__data.append(self.__document[start_pos:formula])
                    start_pos = self.handle_equation(formula)
            elif in_document(formula):
                self.__data.append(self.__document[start_pos:formula])
                start_pos = self.handle_equation(formula)
            elif in_document(comment):
                self.__data.append(self.__document[start_pos:comment])
                start_pos = self.handle_comment(comment)
            else: # only data left
                self.__data.append(self.__document[start_pos:])
                start_pos = end


    def handle_equation(self, start_pos):
        """Parse an equation. The given offset should mark the beginning of this
        equation."""
        # get line and column of `start_pos`
        lnum, pos = get_position(self.__document, start_pos)

        match = EqnParser.State.Equation.value.search(self.__document[start_pos:])
        if not match:
            next_eq = find_anycase(self.__document[start_pos+1:], '<eq')
            closing = find_anycase(self.__document[start_pos:], '</eq>')
            if -1 < next_eq < closing and closing > -1:
                raise ParseException("Unclosed tag found", (lnum, pos))
            raise ParseException("Malformed equation tag found", (lnum, pos))
        end = start_pos + match.span()[1]
        attrs, formula = match.groups()
        if '<eq>' in formula or '<EQ' in formula:
            raise ParseException("Invalid nesting of formulas detected.", (lnum,
                pos))

        # replace HTML entities
        entity = EqnParser.HTML_ENTITY.search(formula)
        while entity:
            formula = re.sub(EqnParser.HTML_ENTITY,
                    html.unescape(entity.groups()[0]), formula)
            entity = EqnParser.HTML_ENTITY.search(formula)
        attrs = attrs.lower()
        displaymath = bool(attrs) and 'env' in attrs and 'displaymath' in attrs
        self.__data.append(((lnum, pos), # let line number count from 0 as well
                displaymath, formula))
        return end


    def handle_comment(self, start_pos):
        match = EqnParser.State.Comment.value.search(self.__document[start_pos:])
        if not match:
            lnum, pos = get_position(self.__document, start_pos)
            # this could be a parser issue, too
            raise ParseException("Improperly formatted comment found", (lnum,
                pos))
        self.__data.append('<!--%s-->' % match.groups()[0])
        return start_pos + match.span()[1] # return end of match

    def get_encoding(self):
        """Return the parsed encoding from the HTML meta data. If none was set,
        UTF-8 is assumed."""
        return self.__encoding

    def get_data(self):
        """Return parsed chunks. These are either strings or tuples with formula
        information, see class documentation."""
        return list(x for x in self.__data if x) # filter empty bits


def gen_id(formula):
    """Generate an id for identifying a formula as an anchor in a document.
    The generated ID is guaranteed to be valid in an XML attribute and it won't
    exceed a certain length.
    If you happen to have a lot of formulas > 150 characters with exactly
    the same content in the document, that'll cause a clash of id's."""
    # for some characters we just use a simple replacement (otherwise the
    # would be lost)
    mapped = {'{':'_', '}':'_', '(':'-', ')':'-', '\\':'.', '^':',', '*':'_'}
    id = []
    prevchar = ''
    for c in formula:
        if prevchar == c:
            continue # avoid multiple same characters
        if c in mapped:
            id.append(mapped[c])
        elif c.isalpha() or c.isdigit():
            id.append(c)
        prevchar = c
    # id's must start with an alphabetical character, so prefix the formula with
    # "formula" to make it a valid html id
    if id and not id[0].isalpha():
        id = ['f', 'o', 'r', 'm', '_'] + id
    if not id: # is empty
        raise ValueError("For the formula '%s' no referencable id could be generated." \
                    % formula)
    return ''.join(id[:150])


class OutsourcedFormulaParser(html.parser.HTMLParser):
    """This HTML parser parses the head and tries to keep it close to the
    original document as possible. As soon as a formula is encountered, only
    the formulas  are parsed. Everything in between and after the them will be
    fully ignored.
    A sample formula would look like:

        <p id="id_as_generated_by_gen_id"><pre>stuff</pre></p>
    """
    def __init__(self):
        self.__head = []
        self.__id = None
        self.__passed_head = False
        self.__equations = collections.OrderedDict()
        super().__init__(convert_charrefs=False)

    def handle_starttag(self, tag, attrs):
        if tag == 'p':
            attrs = dict(attrs)
            if attrs.get('id'):
                self.__id = attrs['id'] # marks beginning of a formula paragraph
                self.__equations[self.__id] = ''
                return
        elif tag == 'body':
            self.__passed_head = True
            self.__head.append('\n<body>\n')
        if not self.__passed_head:
            self.__head.append(self.get_starttag_text())

    def handle_startendtag(self, tag, attrs):
        if self.__id or self.__passed_head: return # skip everything inside a formula
        if attrs:
            self.__head.append('<{} {} />'.format(tag,
                ' '.join(['%s="%s"' % (x[0], x[1]) for x in attrs])))
        else:
            self.__head.append('<%s />' % tag)

    def handle_endtag(self, tag):
        if self.__id: # inside a formula
            if tag == 'p':
                self.__id = None # end formula block
        elif not self.__passed_head:
            formatted = '</%s>' % tag
            self.__head.append(formatted)

    def handle_data(self, data):
        if self.__id:
            self.__equations[self.__id] += data
        elif not self.__passed_head:
            self.__head.append(data)

    def handle_entityref(self, name):
        if self.__id:
            self.__equations[self.__id] += '&%s;' % name
        elif not self.__passed_head:
            self.__head.append('&%s;' % name)

    def handle_charref(self, name):
        if self.__id:
            self.__equations[self.__id] += '&#%s;' % name
        elif not self.__passed_head:
            self.__head.append('&#%s;' % name)

    def handle_comment(self, blah):
        if not self.__passed_head:
            self.__head.append('<!--%s-->' % blah)

    def handle_decl(self, declaration):
        if not self.__passed_head:
            self.__head.append('<!%s>' % declaration)


    def get_head(self):
        """Return a string containing everything before the first formula."""
        return ''.join(self.__head)

    def get_formulas(self):
        """Return an ordered dictionary with id : formula paragraph."""
        return self.__equations

    def error(self, message):
        raise ParseException(message, ('unknown', 'unknown'))

def format_formula_paragraph(formula):
    """Format a formula to appear as if it would have been outsourced into an
    external file."""
    return '<p id="%s"><pre>%s</pre></span></p>\n' % \
            (gen_id(formula), formula)


class HtmlImageFormatter: # ToDo: localisation
    """HtmlImageFormatter(exclusion_filepath='outsourced_formulas.html',
            encoding="UTF-8")
    Format converted formula to be included into the HTML. A typical image
    attribute will contain the path to the image, style information, a CSS class
    to be used in custom CSS style sheets and an alternative text (the LaTeX
    source) for people who disabled images or for blind screen reader users.
    If set, LaTeX formulas exceeding a configurable maximum length will be
    excluded. The image will be a link which leads to the excluded image text.
    The alt attribute is a text-only attribute and e.g. line breaks will be lost
    for screen reader users, so it makes sense for longer formulas to be
    external to be easily readable. Furthermore the alt attribute is limited to
    255 characters, so formula blocks exceeding that limit need to be treated
    differently anyway. If that behavior is not wanted, it can be disabled and
    nothing will be excluded."""

    EXCLUSION_FILE_NAME = 'outsourced-descriptions.html'
    HTML_TEMPLATE_HEAD = ('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"' +
        '\n  "http://www.w3.org/TR/html4/strict.dtd">\n<html>\n<head>\n' +
        '<meta http-equiv="content-type" content="text/html; charset=utf-8"/>' +
        '\n<title>Outsourced Formulas</title>\n</head>\n<!-- ' +
        'DO NOT MODIFY THIS FILE, IT IS AUTOMATICALLY GENERATED -->\n<body>\n')
    def __init__(self, base_path='', link_prefix=None):
        self.__exclude_descriptions = False
        self.__link_prefix = (link_prefix if link_prefix else '')
        self.__base_path = (base_path if base_path else '')
        self.__exclusion_filepath = posixpath.join(self.__base_path, HtmlImageFormatter.EXCLUSION_FILE_NAME)
        if os.path.exists(self.__exclusion_filepath):
            if not os.access(self.__exclusion_filepath, os.W_OK):
                raise OSError('The file %s is not writable!' %
                        self.__exclusion_filepath)
        self.__inline_maxlength=100
        self.__file_head = HtmlImageFormatter.HTML_TEMPLATE_HEAD
        self.__cached_formula_pars = collections.OrderedDict()
        self.__url = ''
        self.initialized = False
        self.initialize() # read already written file, if any
        self.__css = {'inline' : 'inlinemath', 'display' : 'displaymath'}
        self.__replace_nonascii = False

    def set_replace_nonascii(self, flag):
        """If True, non-ascii characters will be replaced through their LaTeX
        command. Note that alphabetical characters will not be replaced, to
        allow easier readibility."""
        self.__replace_nonascii = flag

    def set_max_formula_length(self, length):
        """Set maximum length of a formula before it gets outsourced into a
        separate file."""
        self.__inline_maxlength = length

    def set_inline_math_css_class(self, css):
        """set css class for inline math."""
        self.__css['inline'] = css

    def set_display_math_css_class(self, css):
        """set css class for display math."""
        self.__css['display'] = css

    def set_exclude_long_formulas(self, flag):
        """When set, the LaTeX code of a formula longer than the configured
        maxlength will be excluded and written + linked into a separate file."""
        self.__exclude_descriptions = flag

    def set_url(self, prefix):
        """Set URL prefix which is used as a prefix to the image file in the
        HTML link."""
        self.__url = prefix

    def initialize(self):
        """Initialize the image writer. If a file with already written image
        descriptions exists, this one will be parsed first and new formulas
        appended to it. Otherwise a new file will be written upon ending the
        with-resources block."""
        if self.initialized:
            return
        self.initialized = True
        if not os.path.exists(self.__exclusion_filepath):
            return self
        document = None
        with open(self.__exclusion_filepath, 'r', encoding='UTF-8') as f:
            document = f.read()
        # parse html document:
        parser = OutsourcedFormulaParser()
        parser.feed(document)
        self.__file_head = parser.get_head()
        self.__cached_formula_pars = parser.get_formulas()
        return self

    def __enter__(self):
        return self

    def __exit__(self, useless, unused, not_applicable):
        self.close()

    def close(self):
        """Write back file with excluded image descriptions, if any."""
        def formula2paragraph(frml):
            return '<p id="%s"><pre>%s</pre></p>' % (gen_id(frml), frml)
        if not self.__cached_formula_pars:
            return
        with open(self.__exclusion_filepath, 'w', encoding='utf-8') as f:
            f.write(self.__file_head)
            f.write('\n<hr />\n'.join([formula2paragraph(formula) \
                    for formula in self.__cached_formula_pars.values()]))
            f.write('\n</body>\n</html>\n')

    def get_html_img(self, pos, formula, img_path, displaymath=False):
        """:param pos dictionary containing keys depth, height and width
        :param formula LaTeX alternative text
        :param img_path: path to image
        :param displaymath display or inline math (default False, inline maths)
        :returns a string with the formatted HTML"""
        full_url = img_path
        if self.__url:
            if self.__url.endswith('/'): self.__url = self.__url[:-1]
            full_url = self.__url + '/' + img_path
        # depth is a negative offset
        depth = float(pos['depth']) * -1
        css = (self.__css['display'] if displaymath else self.__css['inline'])
        return ('<img src="{0}" style="vertical-align: {3:.2f}px; margin: 0;" '
                'height="{2[height]:.2f}" width="{2[width]:.2f}" alt="{1}" '
                'class="{4}" />').format(full_url, formula, pos, depth, css)

    def format_excluded(self, pos, formula, img_path, displaymath=False):
        """This method formats a formula and an formula image in HTML and
        additionally writes the formula to an external (configured) file to
        which the image links to. That's useful for blind screen reader users
        who can then easily have a look at the formula.
        :param pos dictionary containing keys depth, height and width
        :param formula LaTeX alternative text
        :param img_path: path to image
        :param displaymath if set to true, image is treated as display math formula (default False)
        :returns string with formatted HTML image which also links to excluded
        formula"""
        shortened = (formula[:100] + '...'  if len(formula) > 100 else formula)
        img = self.get_html_img(pos, shortened, img_path, displaymath)
        identifier = gen_id(formula)
        # write formula out to external file
        if identifier not in self.__cached_formula_pars:
            self.__cached_formula_pars[identifier] = formula
        exclusion_filelink = posixpath.join(self.__link_prefix, self.__exclusion_filepath)
        return '<a href="{}#{}">{}</a>'.format(exclusion_filelink,
                gen_id(formula), img)

    def format(self, pos, formula, img_path, displaymath=False):
        """This method formats a formula. If self.__exclude_descriptions is set
        and the formula igreater than the configured length, the formula will be
        outsourced, otherwise it'll be included in the IMG's alt tag. In either
        case, a string for the current document containing the formatted HTML is returned.
        :param pos dictionary containing keys depth, height and width
        :param formula LaTeX alternative text
        :param img_path: path to image
        :param displaymath whether or not formula is in display math (default: no)
        :returns string with formatted HTML image which also links to excluded
        formula"""
        formula = typesetting.increase_readability(formula,
                self.__replace_nonascii)
        if self.__exclude_descriptions and \
                len(formula) > self.__inline_maxlength:
            return self.format_excluded(pos, formula, img_path, displaymath)
        return self.get_html_img(pos, formula, img_path, displaymath)

def write_html(file, document, formatter):
    """Processed HTML documents are made up of raw HTML chunks which are written
    back unaltered and of processed image. An processed image is a former
    formula converted to an image with additional meta data. This is passed to
    the format function of the supplied formatter and the result is written to
    the given (open) file handle."""
    for chunk in document:
        if isinstance(chunk, dict):
            is_displaymath = chunk['displaymath']
            file.write(formatter.format(chunk['pos'], chunk['formula'],
                    chunk['path'], is_displaymath))
        else:
            file.write(chunk)



import os
import setuptools

with open(os.path.join('gleetex', '__init__.py')) as f:
    global VERSION
    import re
    for line in f.read().split('\n'):
        vers = re.search(r'^VERSION\s+=\s+.(\d+\.\d+\.\d+)', line)
        if vers:
            VERSION = vers.groups()[0]
    if not VERSION:
        class SetupError(Exception):
            pass
        raise SetupError("Error parsing package version")

setuptools.setup(name='GladTeX',
      version=VERSION,
      description="Formula typesetting for the web using LaTeX",
      long_description="""Formula typesetting for the web

This package (and command-line application) allows to create web documents with
properly typeset formulas. It uses the embedded LaTeX formulas from the source
document to place SVG images at the right positions on the web page. For people
not able to see the images (due to a poor internet connection or because of a
disability), the LaTeX formula is preserved in the alt tag of the SVG image.
GladTeX hence combines proper math typesetting on the web with the creation of
accessible scientific documents.""",
      long_description_content_type = "text/markdown",
      author='Sebastian Humenda',
      author_email='shumenda@gmx.de',
      url='https://humenda.github.io/GladTeX',
      packages=['gleetex'],
      entry_points={
       "console_scripts": [
           "gladtex = gleetex.__main__:main"
        ]
      },
      license = "LGPL3.0",
     )



# (c) 2013-2018 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""Top-level API to parse input documents.

The main point of the parsing is to extract formulas from a given input
document, while preserving the remaining formatting.
The returned parsed document structure is highly dependent on the input format
and hence document in their respective functions."""

import enum
import json
import sys

from . import htmlhandling
from . import pandoc
ParseException = htmlhandling.ParseException # re-export for consistent API from outside

class Format(enum.Enum):
    HTML = 0
    # while this is json, we never know what other applications might decide to
    # use json as their intermediate representation ;)
    PANDOCFILTER = 1

    @staticmethod
    def parse(string):
        string = string.lower()
        if string == 'html':
            return Format.HTML
        elif string == 'pandocfilter':
            return Format.PANDOCFILTER
        else:
            raise ValueError("unrecognised format: %s" % string)

def parse_document(doc, fmt):
    """This function parses an input document (string or bytes) with the given
    format specifier. For HTML, the returned "parsed" document is a list of
    chunks, where raw chunks are just plain HTML instructions and data and
    formula chunks are parsed from the '<eq/>' tags.
    If the input document is a pandoc AST, the formulas will be extracted and
    the document is a tuple of (pandoc AST, formulas).

    :param doc  input of bytes or string to parse
    :param fmt  either the enum type `Format` or a string understood by Format.parse
    :return     (encoding, document) (a tuple)"""
    if isinstance(fmt, str):
        fmt = Format.parse(fmt)
    encoding = None
    if fmt == Format.HTML:
        docparser = htmlhandling.EqnParser()
        docparser.feed(doc)
        encoding = docparser.get_encoding()
        encoding = (encoding if encoding else 'utf-8')
        doc = docparser.get_data()
    elif fmt == Format.PANDOCFILTER:
        if isinstance(doc, bytes):
            doc = doc.decode(sys.getdefaultencoding())
        ast = json.loads(doc)
        formulas = pandoc.extract_formulas(ast)
        doc = (ast, formulas) # ← see doc string
    if not encoding:
        encoding = sys.getdefaultencoding()
    return encoding, doc




from . import caching
from . import cachedconverter
from . import htmlhandling
from . import image
from . import pandoc
from . import parser
from . import typesetting

VERSION = '3.1.0'

__all__ = ['caching', 'cachedconverter', 'htmlhandling', 'image',
        'pandoc', 'parser', 'unicode', 'VERSION']



# (c) 2013-2018 Sebastian Humenda
# This code is licenced under the terms of the LGPL-3+, see the file COPYING for
# more details.
"""This module takes care of the actual image creation process.

Each formula is saved as an image, either as PNG or SVG. SVG is advised, since
it is a properly scalable format.
"""

import enum
import os
import re
import shutil
import subprocess
import sys

from .typesetting import LaTeXDocument

DVIPNG_REGEX = re.compile(r"^ depth=(-?\d+) height=(\d+) width=(\d+)")
DVISVGM_DEPTH_REGEX = re.compile(r"^\s*width=.*?pt, height=.*?pt, depth=(.*?)pt")
DVISVGM_SIZE_REGEX = re.compile(r"^\s*graphic size: (.*?)pt x (.*?)pt")

def remove_all(*files):
    """Guarded remove of files (rm -f); no exception is thrown if a file
    couldn't be removed."""
    for file in files:
        try:
            os.remove(file)
        except OSError:
            pass


def proc_call(cmd, cwd=None, install_recommends=True):
    """Execute cmd (list of arguments) as a subprocess. Returned is a tuple with
    stdout and stderr, decoded if not None. If the return value is not equal 0, a
    subprocess error is raised. Timeouts will happen after 20 seconds."""
    with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            bufsize=1, universal_newlines=False, cwd=cwd) as proc:
        data = []
        try:
            data = [d.decode(sys.getdefaultencoding(), errors="surrogateescape")
                    for d in proc.communicate(timeout=20) if d]
            if proc.wait():
                raise subprocess.SubprocessError("Error while executing %s\n%s\n" %
                    (' '.join(cmd), '\n'.join(data)))
        except subprocess.TimeoutExpired as e:
            proc.kill()
            note = 'Subprocess expired with time out: ' + str(cmd) + '\n'
            poll = proc.poll()
            if poll:
                note += str(poll) + '\n'
            if data:
                raise subprocess.SubprocessError(str(data + '\n' + note))
            else:
                raise subprocess.SubprocessError('execution timed out after ' +
                        str(e.args[1]) + ' s: ' + ' '.join(e.args[0]))
        except KeyboardInterrupt as e:
            sys.stderr.write("\nInterrupted; ")
            import traceback
            traceback.print_exc(file=sys.stderr)
        except FileNotFoundError:
            # program missing, try to help
            text = "Command `%s` not found." % cmd[0]
            if install_recommends and shutil.which('dpkg'):
                text += ' Install it using `sudo apt install ' + install_recommends
            else:
                text += ' Install a TeX distribution of your choice, e.g. MikTeX or TeXlive.'
            raise subprocess.SubprocessError(text) from None
        if isinstance(data, list):
            return '\n'.join(data)
        return data

#pylint: disable=too-few-public-methods
class Format(enum.Enum):
    """Chose the image output format."""
    Png = 'png'
    Svg = 'svg'

class Tex2img:
    """Convert a TeX document string into a png file.
    This class interacts with the LaTeX and dvipng sub processes. Upon error
    the methods throw a SubprocessError with all necessary information to fix
    the issue.

    The background of the PNG files will be transparent by default. If you set a
    background colour within the LaTeX document, you need to turn off
    transparency in this converter manually."""
    def __init__(self, fmt, encoding="UTF-8"):
        if not isinstance(fmt, Format):
            raise ValueError("Enumeration of type Format expected."+str(fmt))
        self.__format = fmt
        self.__encoding = encoding
        self.__parsed_data = None
        self.__size = [115, None]
        self.__background = 'transparent'
        self.__keep_latex_source = False

    def set_dpi(self, dpi):
        """Set output resolution for formula images. This has no effect ifthe
        output format is SVG. It will automatically overwrite a font size, if
        set."""
        if not isinstance(dpi, (int, float)):
            raise TypeError("Dpi must be an integer or floating point number")
        self.__size[0] = int(dpi)

    def set_fontsize(self, size):
        """Set font size for formulas. This will be automatically translated
        into a DPI resolution for PNG images and taken literally for SVG
        graphics."""
        if not isinstance(size, (int, float)):
            raise TypeError("Dpi must be an integer or floating point number")
        self.__size[1] = float(size)

    def set_transparency(self, flag):
        """Set whether or not to use background colour information from the DVI
        file. This is only relevant for PNG output and if a background colour
        other than "transparent" is required, in this case this set'r should be
        set to false. It is set to True, resulting in a transparent
        background."""
        self.__background = ('transparent' if flag else 'not transparent')
    def set_keep_latex_source(self, flag):
        """Set whether LaTeX source document should be kept."""
        if not isinstance(flag, bool):
            raise TypeError("boolean object required, got %s." % repr(flag))
        self.__keep_latex_source = flag


    def create_dvi(self, tex_document, dvi_fn):
        """Call LaTeX to produce a dvi file with the given LaTeX document.
        Temporary files will be removed, even in the case of a LaTeX error.
        This method raises a SubprocessError with the helpful part of LaTeX's
        error output."""
        path = os.path.dirname(dvi_fn)
        if path and not os.path.exists(path):
            os.makedirs(path)
        if not path:
            path = os.getcwd()
        new_extension = lambda x: os.path.splitext(dvi_fn)[0] + '.' + x

        if self.__size[1]: # font size in pt
            tex_document.set_fontsize(self.__size[1])
        tex_fn = new_extension('tex')
        aux_fn = new_extension('aux')
        log_fn = new_extension('log')
        cmd = None
        encoding = self.__encoding
        with open(tex_fn, mode='w', encoding=encoding) as tex:
            tex.write(str(tex_document))
        cmd = ['latex', '-halt-on-error', os.path.basename(tex_fn)]
        try:
            proc_call(cmd, cwd=path, install_recommends='texlive-recommended')
        except subprocess.SubprocessError as e:
            remove_all(dvi_fn)
            msg = ''
            if e.args:
                data = self.parse_latex_log(e.args[0])
                if data:
                    msg += data
                else:
                    msg += str(e.args[0])
            raise subprocess.SubprocessError(msg) # propagate subprocess error
        finally:
            if self.__keep_latex_source:
                remove_all(aux_fn, log_fn)
            else:
                remove_all(tex_fn, aux_fn, log_fn)

    def create_image(self, dvi_fn):
        """Create the image containing the formula, using either dvisvgm or
        dvipng."""
        dirname = os.path.dirname(dvi_fn)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

        output_fn = '%s.%s' % (os.path.splitext(dvi_fn)[0], self.__format.value)
        if self.__format == Format.Png:
            dpi = (fontsize2dpi(self.__size[1])  if self.__size[1]
                    else self.__size[0])
            return create_png(dvi_fn, output_fn,dpi,
                    self.__background)
        if not self.__size[1]:
            self.__size[1] = 12 # 12 pt
        return create_svg(dvi_fn, output_fn)

    def convert(self, tex_document, base_name):
        """Convert the given TeX document into an image. The base name is used
        to create the required intermediate files and the resulting file will be
        made of the base_name and the format-specific file extension.
        This function returns the positioning information used in the CSS style
        attribute."""
        if not isinstance(tex_document, LaTeXDocument):
            raise TypeError(("expected object of type typesetting.LaTeXDocument,"
                    " got %s") % type(tex_document))
        dvi = '%s.dvi' % base_name
        try:
            self.create_dvi(tex_document, dvi)
            return self.create_image(dvi)
        except OSError:
            remove_all('%s.%s' % (base_name, self.__format.value))
            raise

    def parse_latex_log(self, logdata):
        """Parse the LaTeX error output and return the relevant part of it."""
        if not logdata:
            return None
        line = None
        for line in logdata.split('\n'):
            if line.startswith('! '):
                line = line[2:]
                break
        if line: # try to remove LaTeX line numbers
            lineno = re.search(r'\s*on input line \d+', line)
            if lineno:
                line = line[:lineno.span()[0]] + line[lineno.span()[1]:]
            return line
        return None

def fontsize2dpi(size_pt):
    """This function calculates the DPI for the resulting image. Depending on
    the font size, a different resolution needs to be used. According to the
    dvipng manual page, the formula is:
    <dpi> = <font_px> * 72.27 / 10 [px * TeXpt/in / TeXpt]"""
    size_px = size_pt * 1.3333333 # and more 3s!
    return size_px * 72.27 / 10

def create_png(dvi_fn, output_name, dpi, background):
    """Create a PNG file from a given dvi file. The side effect is the PNG file
    being written to disk.
    By default, the background of the resulting image is transparent, setting
    any other value will make it use whatever was is set in the DVI file.
    :param dvi_fn       Dvi file name
    :param output_name  Output file name
    :param dpi          Output resolution
    :param background   Background colour (default: transparent)
    :return dimensions for embedding into an HTML document
    :raises ValueError raised whenever dvipng output coudln't be parsed"""
    if not output_name:
        raise ValueError("Empty output_name")
    cmd = ['dvipng', '-q*', '-D', str(dpi)]
    if background == 'transparent':
        cmd += ['-bg', background]
    cmd += ['--height*', '--depth*', '--width*', # print information for embedding
            '-o', output_name, dvi_fn]
    data = None
    try:
        data = proc_call(cmd, install_recommends='dvipng')
    except subprocess.SubprocessError:
        remove_all(output_name)
        raise
    finally:
        remove_all(dvi_fn)
    for line in data.split('\n'):
        found = DVIPNG_REGEX.search(line)
        if found:
            return dict(zip(['depth', 'height', 'width'],
                map(float, found.groups())))
    raise ValueError("Could not parse dvi output: " + repr(data))

def create_svg(dvi_fn, output_name):
    """Create a SVG file from a given dvi file. The side effect is the SVG file
    being written to disk.
    :param dvi_fn       Dvi file name
    :param output_name  Output file name
    :param size         font size in pt
    :return dimensions for embedding into an HTML document
    :raises ValueError raised whenever dvipng output coudln't be parsed"""
    if not output_name:
        raise ValueError("Empty output_name")
    cmd = ['dvisvgm', '--exact', '--no-fonts', '-o', output_name,
            '--bbox=preview', dvi_fn]
    data = None
    try:
        data = proc_call(cmd, install_recommends='texlive-binaries')
    except subprocess.SubprocessError:
        remove_all(output_name)
        raise
    finally:
        remove_all(dvi_fn)
    pos = {}
    for line in data.split('\n'):
        if not pos:
            found = DVISVGM_DEPTH_REGEX.search(line)
            if found:
                # convert from pt to px (assuming 96 dpi)
                pos['depth'] = float(found.groups()[0]) * 1.3333333
        else:
            found = DVISVGM_SIZE_REGEX.search(line)
            if found:
                pos.update(dict(zip(['width', 'height'],
                                    # convert from pt to px (assuming 96 dpi)
                                    (float(v) * 1.3333333 for v in found.groups()))))
                return pos
    raise ValueError("Could not parse dvisvgm output: " + repr(data))
